from typing import Dict

import youtube_dl
from youtube_dl.utils import DownloadError
from werkzeug.exceptions import BadRequest


def get_url_metadata(url: str) -> Dict[str, str]:
    """
    Return the metada associated to an URL.

    Return format is a dict:
    ```
    {
        "artist": <string>,
        "duration": <int>,
        "title": <string>,
        "url": <string>
    }
    ```
    """
    with youtube_dl.YoutubeDL({}) as ydl:
        try:
            metadata = ydl.extract_info(url, download=False)
        except DownloadError:
            raise BadRequest("Unrecognized URL")
        return {
            "artist": metadata["uploader"],
            "duration": metadata["duration"],
            "title": metadata["title"],
            "url": metadata["webpage_url"],
        }



from enum import Enum
from subprocess import Popen
from typing import Dict, List, Optional
import json
import select
import socket

from flask import current_app, _app_ctx_stack, Flask


State = Enum("State", ("PLAYING", "PAUSED", "STOPPED"))


class InvalidPlayerState(ValueError):
    pass


class Player:
    """
    Control mpv using IPC.

    This is a Flask extension.
    """

    _command_id = 0
    _state = State.STOPPED

    def __init__(self, app: Flask = None) -> None:
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        self.app = app

        app.config.setdefault("MPV_IPC_SOCKET", "/tmp/mpv-socket")
        app.teardown_appcontext(self.teardown)

        @app.before_first_request
        def init_mpv() -> None:
            self.mpv = Popen(
                [
                    "mpv",
                    "--idle",
                    "--no-terminal",
                    f"--input-ipc-server={app.config['MPV_IPC_SOCKET']}",
                ]
            )

    def teardown(self, exception: Optional[Exception]) -> None:
        ctx = _app_ctx_stack.top

        if hasattr(ctx, "mpv_socket"):
            ctx.mpv_socket.close()

    def add_item(self, url: str) -> None:
        """Add a new item to the playlist.

        If the player was stopped, play the music.
        """
        if self._state == State.STOPPED:
            self._send_command("loadfile", url, "append-play")
            self._state = State.PLAYING
        else:
            self._send_command("loadfile", url, "append")

    def get_current_item_title(self) -> Optional[str]:
        """Get the current playing item's title."""
        response = self._send_command("get_property", "metadata")
        if response is not None:
            return response["data"]["title"]

        return None

    def get_state(self) -> State:
        return self._state

    def play(self) -> None:
        """Play the current playlist."""
        if self._state == State.PLAYING:
            return
        elif self._state == State.STOPPED:
            raise InvalidPlayerState("The player is stopped. Add an item to play it.")
        else:
            self._send_command("cycle", "pause")

        self._state = State.PLAYING

    def pause(self) -> None:
        """Pause the current playlist."""
        if self._state == State.PAUSED or self._state == State.STOPPED:
            return

        self._send_command("cycle", "pause")
        self._state = State.PAUSED

    @property
    def _connection(self) -> socket.socket:
        ctx = _app_ctx_stack.top
        if ctx is None:
            raise RuntimeError("No app context available")

        if not hasattr(ctx, "mpv_socket"):
            ctx.mpv_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            ctx.mpv_socket.connect(current_app.config["MPV_IPC_SOCKET"])

        return ctx.mpv_socket

    def _get_command_id(self) -> int:
        self._command_id += 1
        return self._command_id

    def _send_command(self, *command: str) -> Optional[Dict]:
        """Send a command to MPV and return the response."""
        command_id = self._get_command_id()
        msg = {"command": command, "request_id": command_id}
        self.app.logger.debug("Sending MPV commanv %s", msg)

        data = json.dumps(msg).encode("utf8") + b"\n"
        self._connection.sendall(data)

        return self._read_response(command_id)

    def _read_response(self, command_id: int) -> Optional[Dict]:
        data = b""
        while True:
            r, _, _ = select.select([self._connection], [], [], 1)
            if not r:
                # timeout
                return None

            data += self._connection.recv(1024)

            newline = data.find(b"\n")
            if newline == -1:
                next

            response = data[:newline]
            data = data[newline:]

            msg = json.loads(response.decode("utf8"))
            self.app.logger.debug("Received MPV response %s", msg)
            if msg.get("request_id", -1) == command_id:
                return msg



from flask_restplus import Api

from .items import ns as items_ns
from .player import ns as player_ns

api = Api(
    version="0.1",
    title="Escarpolette",
    description="Manage your party's playlist without friction",
)

api.add_namespace(items_ns)
api.add_namespace(player_ns)



from escarpolette.extensions import db
from datetime import datetime
from sqlalchemy import text


class Item(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    created_at = db.Column(
        db.DateTime,
        default=datetime.now,
        server_default=text("datetime()"),
        nullable=False,
    )
    artist = db.Column(db.String(255))
    duration = db.Column(db.Integer)
    played = db.Column(db.Boolean, default=False, server_default=text("FALSE"))
    title = db.Column(db.String(255))
    url = db.Column(db.String(255), unique=True)



from flask import Flask


class Default:
    # Server
    HOST = "127.0.0.1"

    # Database
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    def __init__(self, app: Flask):
        self.SQLALCHEMY_DATABASE_URI = f"sqlite:///{app.instance_path}/db.sqlite"



from typing import Dict

from flask import request
from flask_restplus import Namespace, Resource, fields
from werkzeug.exceptions import BadRequest

from escarpolette.extensions import db, player
from escarpolette.models import Item
from escarpolette.tools import get_url_metadata

ns = Namespace("items", description="Manage the playlist's items")

item = ns.model(
    "Item",
    {
        "artist": fields.String(readonly=True),
        "duration": fields.Integer(readonly=True),
        "title": fields.String(readonly=True),
        "url": fields.Url(
            absolute=True,
            example="https://www.youtube.com/watch?v=bpA6fAz_r04",
            required=True,
        ),
    },
)

playlist = ns.model(
    "Playlist", {"idx": fields.Integer, "items": fields.List(fields.Nested(item))}
)


@ns.route("/")
class Items(Resource):
    @ns.marshal_list_with(playlist)
    def get(self) -> Dict:
        playlist = []
        playing_idx = 0

        for item in Item.query.order_by(Item.created_at).all():
            playlist.append(
                {
                    "artist": item.artist,
                    "duration": item.duration,
                    "title": item.title,
                    "url": item.url,
                }
            )
            if item.played:
                playing_idx += 1

        data = {"items": playlist, "idx": playing_idx}
        return data

    @ns.expect(item)
    @ns.marshal_with(item, code=201)
    def post(self) -> Dict:
        data = request.json

        if data is None:
            raise BadRequest("Missing data")

        metadata = get_url_metadata(data["url"])
        item = Item(**metadata)
        db.session.add(item)
        db.session.flush()

        player.add_item(item.url)

        db.session.commit()

        return metadata



from flask import Flask
from flask_cors import CORS
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy
from werkzeug.local import LocalProxy

from escarpolette.player import Player


db = SQLAlchemy()
player = Player()


def init_app(app: Flask):
    CORS(app)
    Migrate(app, db)
    db.init_app(app)
    player.init_app(app)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['escarpolette',
 'escarpolette.admin',
 'escarpolette.api',
 'escarpolette.models']

package_data = \
{'': ['*']}

install_requires = \
['flask-cors>=3.0,<4.0',
 'flask-migrate>=2.5,<3.0',
 'flask-restplus>=0.12.1,<0.13.0',
 'flask-sqlalchemy>=2.3,<3.0',
 'flask>=1.0,<2.0',
 'sqlalchemy>=1.2,<2.0',
 'youtube-dl>=2019.1,<2020.0']

setup_kwargs = {
    'name': 'escarpolette',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Alexandre Morignot',
    'author_email': 'erdnaxeli@cervoi.se',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from flask import Flask

from escarpolette.api import api
from escarpolette.settings import Default as DefaultSettings
from escarpolette import extensions

app = Flask(__name__, instance_relative_config=True)
app.config.from_object(DefaultSettings(app))
app.config.from_pyfile("application.cfg", silent=True)

extensions.init_app(app)
api.init_app(app)



from typing import Dict

from flask import request
from flask_restplus import Namespace, Resource, fields
from werkzeug.exceptions import BadRequest

from escarpolette.extensions import player
from escarpolette.player import State

ns = Namespace("player", description="Manage the player state")

player_model = ns.model(
    "Player", {"state": fields.String(enum=[x.name for x in State], required=True)}
)


@ns.route("/")
class Player(Resource):
    @ns.marshal_list_with(player_model)
    def get(self) -> Dict:
        data = {"state": player.get_state().name}
        return data

    @ns.expect(player_model)
    @ns.response(200, "Player state changed")
    @ns.response(400, "Invalid state")
    def patch(self) -> None:
        data = request.get_json()
        if data["state"] == State.PLAYING:
            player.play()
        elif data["state"] == State.PAUSED:
            player.pause()
        else:
            raise BadRequest(f"The state f{data['state']} cannot be set.")
